#include <iostream>

#include <cstdint>

// When start and end are compile-time constants, these functions compile into
// efficient code. Make them templates?
inline uint64_t bits(const uint8_t* data, int start, int end) __attribute__((always_inline));
inline uint64_t bits(uint64_t value, int start, int end) __attribute__((always_inline));
inline void set_bits(uint8_t* data, int start, int end, uint64_t value) __attribute__((always_inline));
inline uint64_t set_bits(uint64_t value, int start, int end, uint64_t bits) __attribute__((always_inline));

// Treats data as a little-endian number and extracts bits [start:end]
// (inclusive) from it
inline uint64_t bits(const uint8_t* data, int start, int end) {
  int is = start / 8; // index of the start byte
  int bs = start % 8; // first bit in the start byte
  int ie = end   / 8; // index of the end byte
  int be = end   % 8; // last bit in the ending byte
  unsigned x = data[ie] & 0xFF >> 7 - be;
  if (ie == is) return x >> bs;
  for (int i = ie; --i > is;) x = x << 8 | data[i];
  return x << 8 - bs | data[is] >> bs;
};

// Treats data as a little-endian number and sets bits [start:end] (inclusive)
// to value
// XXX: gcc-13.2.1 doesn't want to inline this function without the declaration above.
inline void set_bits(uint8_t* data, int start, int end, uint64_t value) {
  int is = start / 8;
  int bs = start % 8;
  int ie = end   / 8;
  int be = end   % 8;

  if (ie == is) {
    uint8_t mask = ~(0xFF << (be - bs)) << bs;
    data[ie] = data[ie] & ~mask | value & mask;
  } else {
    data[is] = data[is] & 0xFF >> 8 - bs | value << bs;
    value >>= 8 - bs;
    for (int i = is + 1; i < ie; ++i) {
      data[i] = value;
      value >>= 8;
    };
    uint8_t mask = 0xFF >> 7 - be;
    data[ie] = value & mask | data[ie] & ~mask;
  };
};

// Extracts bits [start:end] (inclusive) from value
inline uint64_t bits(uint64_t value, int start, int end) {
  return (value & (1ULL << end + 1) - 1) >> start;
};

// Sets bits [start:end] (inclusive) in value to bits
inline uint64_t set_bits(uint64_t value, int start, int end, uint64_t bits) {
  uint64_t mask = ~(~1ULL << end - start) << start;
  return value & ~mask | bits << start & mask;
};

class TDCHit {
  public:
    TDCHit() {
      std::fill(std::begin(data), std::end(data), 0);
    };

    TDCHit(uint32_t header, uint32_t measurement, uint32_t extended_time) {
      std::fill(std::begin(data), std::end(data), 0);
      SetGEO(bits(header, 0, 4));
      SetEvent(bits(header, 5, 26));
      SetMeasurement(bits(measurement, 0, 20));
      SetChannel(bits(measurement, 21, 25));
      SetTrailing(bits(measurement, 26, 26));
      SetExtendedTime(bits(extended_time, 0, 26));
    };

    uint32_t GetMeasurement() const {
      return bits(data, 0, 20);
    };

    void SetMeasurement(uint32_t measurement) {
      set_bits(data, 0, 20, measurement);
    };

    uint32_t GetExtendedTime() const {
      return bits(data, 21, 47);
    };

    void SetExtendedTime(uint32_t time) {
      set_bits(data, 21, 47, time);
    };

    uint64_t GetValue() const {
      return bits(data, 0, 47);
    };

    void SetValue(uint64_t value) {
      set_bits(data, 0, 47, value);
    };

    uint32_t GetEvent() const {
      return bits(data, 48, 69);
    };

    void SetEvent(uint32_t event) {
      set_bits(data, 48, 69, event);
    };

    uint16_t GetChannel() const {
      return bits(data, 72, 76);
    };

    void SetChannel(uint8_t channel) {
      set_bits(data, 72, 76, channel);
    };

    bool GetTrailing() const {
      return bits(data, 77, 77);
    };

    void SetTrailing(bool trailing) {
      set_bits(data, 77, 77, trailing);
    };

    uint8_t GetGEO() const {
      return bits(data, 80, 84);
    };

    void SetGEO(uint8_t geo) {
      set_bits(data, 80, 84, geo);
    };

    void Print(std::ostream& output = std::cout) const {
      output << "measurement = " << GetMeasurement() << std::endl;
      output << "extended_time = " << GetExtendedTime() << std::endl;
      output << "value = " << GetValue() << std::endl;
      output << "event = " << GetEvent() << std::endl;
      output << "channel = " << GetChannel() << std::endl;
      output << "trailing = " << GetTrailing() << std::endl;
      output << "geo = " << GetGEO() << std::endl;
    };

  private:
    /* bits     width   field
     *  0 - 20  21      measurement
     * 21 - 47  27      extended time
     * 48 - 69  22      event
     * 72 - 76   5      channel
     * 77 - 77   1      trailing
     * 80 - 84   5      geo

       01234567 
     0 mmmmmmmm         0 = zero
     1 mmmmmmmm         m = measurement
     2 mmmmmTTT         c = channel
     3 TTTTTTTT         t = trailing
     4 TTTTTTTT         C = event
     5 TTTTTTTT         T = extended time
     6 CCCCCCCC
     7 CCCCCCCC
     8 CCCCCC00
     9 ccccct00
    10 ggggg000
    */
    uint8_t data[11];
};

class QDCHit {
  public:
    QDCHit() {
      std::fill(std::begin(data), std::end(data), 0);
    };

    QDCHit(
        uint32_t header,
        uint32_t packet,
        uint32_t trailer
    ) {
      SetCrate(bits(header, 16, 23));
      SetValue(bits(packet, 0, 11));
      SetOverflow(bits(packet, 12, 12));
      SetUnderThreshold(bits(packet, 13, 13));
      SetChannel(bits(packet, 16, 20));
      SetGEO(bits(packet, 27, 31));
      SetEvent(bits(trailer, 0, 23));
    };

    uint16_t GetValue() const {
      return bits(data, 0, 11);
    };

    void SetValue(uint16_t value) {
      set_bits(data, 0, 11, value);
    };

    bool GetUnderThreshold() const {
      return bits(data, 12, 12);
    };

    void SetUnderThreshold(bool ut) {
      set_bits(data, 12, 12, ut);
    };

    bool GetOverflow() const {
      return bits(data, 13, 13);
    };

    void SetOverflow(bool overflow) {
      set_bits(data, 13, 13, overflow);
    };

    uint8_t GetGEO() const {
      return bits(data, 14, 18);
    };

    void SetGEO(uint8_t geo) {
      set_bits(data, 13, 13, geo);
    };

    uint8_t GetChannel() const {
      return bits(data, 19, 23);
    };

    void SetChannel(uint8_t channel) {
      set_bits(data, 19, 23, channel);
    };

    uint8_t GetCrate() const {
      return bits(data, 24, 31);
    };

    void SetCrate(uint8_t crate) {
      set_bits(data, 24, 31, crate);
    };

    uint32_t GetEvent() const {
      return bits(data, 32, 55);
    };

    void SetEvent(uint32_t event) {
      set_bits(data, 32, 55, event);
    };

    void Print(std::ostream& output = std::cout) const {
      output << "value = " << GetValue() << std::endl;
      output << "under_threshold = " << GetUnderThreshold() << std::endl;
      output << "overflow = " << GetOverflow() << std::endl;
      output << "geo = " << GetGEO() << std::endl;
      output << "channel = " << GetChannel() << std::endl;
      output << "crate = " << GetCrate() << std::endl;
      output << "event = " << GetEvent() << std::endl;
    };

  private:
    /* bits     width   field
     *  0 - 11   12     value
     * 12 - 12   1      under threshold
     * 13 - 13   1      overflow
     * 14 - 18   5      geo
     * 19 - 23   5      channel
     * 24 - 31   8      crate
     * 32 - 55   24     event

       01234567
     0 vvvvvvvv         v = value
     1 vvvvuogg         u = under threshold
     2 gggccccc         o = overflow
     3 CCCCCCCC         g = geo
     4 eeeeeeee         c = channel
     5 eeeeeeee         C = crate
     6 eeeeeeee         e = event
     */

    uint8_t data[7];
};
